package ca.ualberta.cs.poker.free.tournament;

import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.*;

import ca.ualberta.cs.poker.free.server.PokerServer;

/*
 * This class contains the necessary information for a match between two competitors to
 * be run. Its implements Runnable and calling run() will run the entire match.
 * 
 * @author Christian Smith
 * 
 */
public class Match implements Runnable {
	
	private String ip1;
	private String ip2;
	private String comp1;
	private String comp2;
	private int compNum1;
	private int compNum2;
	private int series;
	private int match;
	private int timePerHand;
	private String type;
	private String cardFile;
	private String logFile;
	private String resultFile;
	
	
	public Match( String _comp1, String _comp2, String _ip1, String _ip2, int _series, int _match, String _type, int _timePerHand) {
		
		type = _type;
		ip1 = _ip1;
		ip2 = _ip2;
		comp1 = _comp1;
		comp2 = _comp2;
		series = _series;
		match = _match;
		timePerHand = _timePerHand;
		
		if (  ! (type.equals("SE")  || (type.equals("RR"))  )) {
			System.err.println( "The match type must be either SE (Single Elimination) or (Round Robin)." );
			System.exit(0);
		}
		
		if ( type == "RR" ) {
			cardFile = "data/cards/bankroll" + comp1 + comp2 + "match" + match + ".crd";
			logFile = "data/logs/bankroll" + comp1 + comp2 + "match" + match + ".log";
			resultFile = "data/results/bankroll" + comp1 + comp2 + "match" + match + ".res";
		}
		if ( type == "SE" ) {
			cardFile = "data/cards/seseries" + series + "match" + match + ".crd";
			logFile = "data/logs/seseries" + series + "match" + match + ".log";
			resultFile = "data/results/seseries" + series + "match" + match + ".res";
			
		}
		
		
		
		
	}
	
public Match( int _comp1, int _comp2, String _ip1, String _ip2, int _series, int _match, String _type, int _timePerHand ) {
		
		type = _type;
		ip1 = _ip1;
		ip2 = _ip2;
		compNum1 = _comp1;
		compNum2 = _comp2;
		comp1 = "comp" + _comp1;
		comp2 = "comp" + _comp2;
		series = _series;
		match = _match;
		timePerHand = _timePerHand;
		
		if (  ! (type.equals("SE")  || (type.equals("RR"))  )) {
			System.err.println( "The match type must be either SE (Single Elimination) or (Round Robin)." );
			System.exit(0);
		}
		
		if ( type == "RR" ) {
			cardFile = "data/cards/bankroll" + comp1 + comp2 + "match" + match + ".crd";
			logFile = "data/logs/bankroll" + comp1 + comp2 + "match" + match + ".log";
			resultFile = "data/results/bankroll" + comp1 + comp2 + "match" + match + ".res";
		}
		if ( type == "SE" ) {
			cardFile = "data/cards/seseries" + series + "match" + match + ".crd";
			logFile = "data/logs/seseries" + series + "match" + match + ".log";
			resultFile = "data/results/seseries" + series + "match" + match + ".res";
			
		}	
	}
	
/*
 * FlipRRFilePaths
 * 
 * In the case we want to run the same cards with the players seated in opposet
 * spots, this will allow the card file to be the same, but the log and result
 * file to reflect the change in seating.
 * 
 */
	public void flipRRFilePaths( ) {
		cardFile = "data/cards/bankroll" + comp2 + comp1 + "match" + match + ".crd";
		logFile = "data/logs/bankroll" + comp1 + comp2 + "match" + match + "rev.log";
		resultFile = "data/results/bankroll" + comp1 + comp2 + "match" + match + "rev.res";
	}

	
        public boolean isComplete(){
            return (new File(getResultFile())).exists();
        }
        
        
	/*
	 * When comparing two matches, compare the series and match number in Single Elimination
	 * 
	 * In Round Robin check the competitors seat position, via the result file
	 * 
	 */
	public boolean equals( Object obj ) {
		
		if( obj instanceof Match ) {
			
			if ( ((Match)obj).getType().equals("SE" ) ) {
	
				if (  (((Match)obj).getSeries() == series) && (((Match)obj).getMatch() == match) ) {
					return true;
				}
				else
					return false;			
			}
			if ( ((Match)obj).getType().equals("RR" ) ) {
			
//				if ( ((Match)obj).getCompNum1() == this.getCompNum1() &&
//					((Match)obj).getCompNum1() == this.getCompNum1() &&
//						((Match)obj).getMatch() == this.getMatch() ) {
				if ( ((Match)obj).getResultFile().equals( this.getResultFile() ) ) {
					return true;
				}
				else
					return false;
			}
			
			
		}
		else
			return false;
		
		// should be unreachable
		return false;
		
	}
	

	/*
	 * Getters and Setters
	 */
	public String getIp1() {
		return ip1;
	}

	public void setIp1(String ip1) {
		this.ip1 = ip1;
	}

	public String getIp2() {
		return ip2;
	}

	public void setIp2(String ip2) {
		this.ip2 = ip2;
	}

	public int getMatch() {
		return match;
	}

	public void setMatch(int match) {
		this.match = match;
	}

	public int getSeries() {
		return series;
	}

	public void setSeries(int series) {
		this.series = series;
	}
	
	public String toString() {
		if ( type.compareTo( "SE" ) == 0 ) {
			return "*** Series: " + series  + " Match: " + match;
		}
		
		else if ( type.compareTo( "RR" ) == 0 ) {
			return "*** Competitor " + compNum1 + " vs " + compNum2 + " Match " + match;
		}
		else
			return "Unknown Match type";
		
	}

	public String getComp1() {
		return comp1;
	}

	public void setComp1(String comp1) {
		this.comp1 = comp1;
	}

	public String getComp2() {
		return comp2;
	}

	public void setComp2(String comp2) {
		this.comp2 = comp2;
	}

	public String getCardFile() {
		return cardFile;
	}

	public String getLogFile() {
		return logFile;
	}

	public String getResultFile() {
		return resultFile;
	}

	/*
	 * This is the threaded version of runMatch. This method will launch a new thread
	 * that will use this class's members to determine which machines need to be cleaned
	 * (remove the directories the client will run from), start the server on its own thread
	 * and launch the clients via ssh. This is now threaded so that if a match fails, we have
	 * a single point to fall back from and try to re-attempt the match.
	 *
	 * 
	 */
	public void run() {
		
		// this method used to be parameterized to take in Match
		Match m = this;
		
		String comp1 = m.getComp1();
		String comp2 = m.getComp2();
		
		String client1 = m.getIp1();
		String client2 = m.getIp2();
		
		String[] cmd1 = new String[3];
		String[] cmd2 = new String[3];
		String user1 = "pokert";
		String user2 = "pokert";
		
		
		String bellis = "129.128.23.142";
		String homer = "129.128.23.106";
		
		String redirOutErr = " > ../out.txt 2> ../err.txt ";
		
		//String user1 = "chsmith";
		//String user2 = "christian";
		//String serverIP = bellis;
		
		String serverIP = "10.2.99.1";
		
		String login1 = user1 + "@" + client1;
		String login2 = user2 + "@" + client2;
		
		// the ' is so cygwin will interpret the two commands in one go
		String command1 = "'tar -xf " + comp1 + ".tar" + redirOutErr + ";" + "cd " + comp1 + "; " +  "./startme.bat " + serverIP + " ";// + port + "'";
		String command2 = "'tar -xf " + comp2 + ".tar" + redirOutErr + ";" + "cd " + comp2 + "; " + "./startme.bat " + serverIP + " ";// + port + "'";
		
		
		// cmd.exe /C is so the directory slashes are interpreted correctly
		cmd1[0] = "cmd.exe ";
		cmd1[1] = "/C ";
		cmd1[2] = "ssh " + login1 + " " + command1;

		cmd2[0] = "cmd.exe ";
		cmd2[1] = "/C ";
		cmd2[2] = "ssh " + login2 + " " + command2;
		
		// Before we copy the client and start the clients, lets clean the machine
		String taskKillCommand1 = "ssh " + login1 + " " + cmd1[0] +  cmd1[1] + "taskkill.exe /F /FI \"USERNAME eq " + user1 + "\"";
		String taskKillCommand2 = "ssh " + login2 + " " + cmd2[0] +  cmd2[1] + "taskkill.exe /F /FI \"USERNAME eq " + user2 + "\"";
		
		String directory1 = comp1; 
		String directory2 = comp2;
		
		String taskCleanCommand1 = "ssh " + login1 + " " + "'rm -rf ~/" + directory1 + "'";
		String taskCleanCommand2 = "ssh " + login2 + " " + "'rm -rf ~/" + directory2 + "'";
		
			
		
		
		// taskkill all the processes, and clear the directories
		// function'ified for easy on/off
		cleanMachines( taskKillCommand1, taskKillCommand2, taskCleanCommand1, taskCleanCommand2);
		

		// Copy over the files needed to run each client
		// must be in directories relative to where RunTournament is being run
		// scp -r localfiles user@host:dir
		String scpCommand1 = "scp -r " + directory1 + ".tar" + " " + login1 + ":~";
		String scpCommand2 = "scp -r " + directory2 + ".tar" + " " + login2 + ":~";



		
		// copy both clients to the client machines with scp
		try {
			
			System.out.println("\n[********** Begin Copy ********]");
			System.out.println(scpCommand1);
			System.out.println(scpCommand2);
			
			
			Process scpChild1 = Runtime.getRuntime().exec(scpCommand1);
			Process scpChild2 = Runtime.getRuntime().exec(scpCommand2);
			
			
			scpChild1.waitFor();
			scpChild2.waitFor();
			
			if ( scpChild1.exitValue() != 0 && scpChild2.exitValue() != 0 ) {
				System.out.println( "Got a non-zero exit code from scp");
				
			}
			  
		    System.out.println("[********** End Copy ********]" );
		  

		} catch (IOException e1) {
			// TODO Auto-generated catch block
			System.err.println( "Error copying the client files");
			e1.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		
		
		// On to starting the server, then start clients
		try {			
			
			System.out.println( "\n[********** Server Starting *********]" );
			
			// the server will write out these 3 files
			FileReader cardReader = new FileReader( m.getCardFile() );
			BufferedWriter logWriter = new BufferedWriter( new FileWriter( m.getLogFile()));
						
			
			// log file for each match, to check if something goes wrong, stdout isnt very useful
			//FileWriter log = new FileWriter( "data/logs/series" + m.getSeries() + "match" + m.getMatch() );
			//FileWriter log = new FileWriter( "data/logs/" + m.getComp1() + "vs" + m.getComp2() + "match" + m.getMatch() + ".log" );
			BufferedWriter log = new BufferedWriter(new FileWriter("data/logs/serverlog" + m.getComp1() + "vs" + m.getComp2() + "match" + m.getMatch() + ".log"));
			
			//log.write("Starting server with " + client1.toString() + " " + client2.toString());
			
			PokerServer server = new PokerServer( client1, client2, timePerHand, cardReader, logWriter, m.getResultFile());
			
			Thread serverThread = new Thread(server);
			serverThread.start();

			// launch server, wait for clients
			while (server.socket == null) {
				Thread.currentThread().sleep(1000);
				System.out.println("Server sleeping one second.");
			}

			log.write("Server reports port: " + server.socket.getLocalPort());
			
			// if port is null, something went horribly wrong
			if ( server.socket.getLocalPort() == 0) {
				log.write("Error getting port from server" );
				//System.err.println( "Error getting port from server" );
				//System.exit(-1);
				return;
			}
			
			
			System.out.println( "\n[********* Clients Starting *********]" );

			
			//launch each client with ssh
			

			String client1Exec = cmd1[2] + server.socket.getLocalPort() + redirOutErr + "'";
			String client2Exec = cmd2[2] + server.socket.getLocalPort() + redirOutErr + "'";
			
                        System.out.println("New code using out and err redirect");
			System.out.println(client1Exec);
			System.out.println(client2Exec);
			
			Process clientChild1 = Runtime.getRuntime().exec(client1Exec);
			Process clientChild2 = Runtime.getRuntime().exec(client2Exec);

			
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("Error reading from child stream");
		} catch (InterruptedException e) {
			
			e.printStackTrace();
		}
	}
	
public static void cleanMachines( String taskKillCommand1, String taskKillCommand2, String taskCleanCommand1, String taskCleanCommand2 ) {
		
		System.out.println( "\n[******** Cleaning Begin ********]");
		//System.out.println( taskKillCommand1 );
		//System.out.println( taskKillCommand2 );
		System.out.println( taskCleanCommand1 );
		System.out.println( taskCleanCommand2 );
		
		
		// TASKILL everything by the user
		// blow away the directory we're going to be adding
		try {
			
			Process clientCleaner1 = Runtime.getRuntime().exec(taskKillCommand1);
			Process clientCleaner2 = Runtime.getRuntime().exec(taskKillCommand2);
			
			//clientCleaner1.waitFor();
			//clientCleaner2.waitFor();
			
			Process clientCleaner3 = Runtime.getRuntime().exec(taskCleanCommand1);
			Process clientCleaner4 = Runtime.getRuntime().exec(taskCleanCommand2);
			
			clientCleaner3.waitFor();
			clientCleaner4.waitFor();
			
			if ( clientCleaner3.exitValue() != 0 && clientCleaner4.exitValue() != 0 ) {
				System.out.println( "Cleaning commands returned a non-zero value");
			}
			
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			System.err.println( "Failure trying to clean client machines");
			e1.printStackTrace();
			
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	
		System.out.println( "[******** End Cleaning ********]");
	}

public int getCompNum1() {
	return compNum1;
}

public int getCompNum2() {
	return compNum2;
}

public String getType() {
	return type;
}

public void setCardFile(String cardFile) {
	this.cardFile = cardFile;
}
	
}
